// Copyright 2007 Theodor Anschtz
// Requires Avisynth 2.5 source code to compile
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
//
// Version history:
//
// 05-26-2007   Ver. 0.1     Initial version
// T. Anschtz
//
// 06-29-2007   Ver. 0.2     Added E, S, and N modes in frameOverrides parameter
// T. Anschtz
//
// 09-09-2007   Ver. 0.3     Moved calls to child->GetFrame into getMinMax and isSceneStart
// T. Anschtz               since having them in in Autolevels::GetFrame made caching pointless.
//                           Changed sceneChgThreshDefault to 20.
//
// 11-04-2010   Ver. 0.4     + removed tabs in source code
// Jim Battle                + comments in English (thanks to google translate)
//                           + limit averaging window to the first scene change
//                             after current frame
//                           + added support for YUY2
//                           + added support for interleaved RGB
//                             the luma mapping is pretty rough; if you want something
//                             more accurate, build the chain yourself:
//                                source = <some rgb32 source>
//                                leveled = source.converttoyv12().autolevels().converttorgb32()
//                           + speedup?: don't compute ymin/ymax in inner loop
//                           + complain if the format isn't supported
//                           + complain about bad exludeframes specifier ([^ENS])
//                           + added "matrix" option for RGB luma calculation
//                           + added boolean "debug" option; text drawing
//                             courtesy of DDigit routines from StainlessS

#include "stdafx.h"
#include "math.h"
#include "hash_map"
#include "vector"
#include "string"
#include "sstream"

#include "avisynth.h"
#include "DDigit.h"    // DrawString()
#include "stdio.h"     // for sprintf

using namespace std;
using namespace stdext;

typedef pair <int, int> IntPair;
typedef pair <int, IntPair> MinMaxElement;
typedef pair <int, boolean> SceneChgElement;
typedef unsigned int unit32;

const int filterRadiusDefault = 5;
const int sceneChgThreshDefault = 20;
const boolean debugDefault = false;

class Autolevels : public GenericVideoFilter
{
    private:
        // filter parameters:
        hash_map <int, boolean> sceneChgOverrides;
        vector<IntPair> excludeFrames;   // numbers of frames (from / to) that are to remain untouched
        int filterRadius;
        int sceneChgThresh;

        // filter parameters not yet exposed:
        enum { luma_Rec709, luma_Rec601, luma_avg } greyscale_mode;
        boolean debug;

        // working variables:
        hash_map <int, IntPair> MinMaxCache;   // holds the ymininner- and ymaxinner- values per frame
        char msgbuf[64];

        // internal routines:
        IntPair getMinMax(int frameno, IScriptEnvironment* env);
        boolean isSceneStart(int frameno, IScriptEnvironment* env);

        void RemapPlanarY        (PVideoFrame &frame, int yLUT[]);
        void RemapYUY2           (PVideoFrame &frame, int yLUT[]);
        void RemapInterleavedRGB (PVideoFrame &frame, int yLUT[]);

        void GetAvgLumaPlanarY        (PVideoFrame &frame, int *yhisto);
        void GetAvgLumaYUY2           (PVideoFrame &frame, int *yhisto);
        void GetAvgLumaInterleavedRGB (PVideoFrame &frame, int *yhisto);

    public:
        Autolevels(PClip _child, int filterRadius, int sceneChgThresh, string frmOverrides, const char *matrix, boolean debug, IScriptEnvironment *env);
        PVideoFrame __stdcall GetFrame(int frameno, IScriptEnvironment* env);
};

Autolevels::Autolevels(PClip _child, int filterRadius, int sceneChgThresh, string frmOverrides, const char *matrix, boolean debug, IScriptEnvironment *env)
    : GenericVideoFilter(_child)
{
    this->filterRadius = filterRadius;
    this->sceneChgThresh = sceneChgThresh;
    this->debug = debug;

    boolean supported = vi.IsYUV() || (vi.IsRGB() && !vi.IsPlanar());
    if (!supported) {
        env->ThrowError("autolevels: planar RGB formats are not currently supported");
    }

    // fill excludeFrames and sceneChgOverrides
    for (int i=0; i<(int)frmOverrides.length(); i++) {
        if (frmOverrides[i] == ',')
            frmOverrides[i] = ' ';
    }

    stringstream ss(frmOverrides);
    string range;
    while (ss >> range) {
        size_t dashPos = range.find_first_of('-');
        int startFrm, endFrm;
        if (dashPos != string::npos) {   // specified area (from - to)
            if (!(istringstream(range.substr(1, dashPos-1)) >> startFrm))
                continue;
            if (!(istringstream(range.substr(dashPos+1)) >> endFrm))
                continue;
        }
        else {   // show unique image ID
            if (!(istringstream(range.substr(1)) >> startFrm))
                continue;
            endFrm = startFrm;
        }
        switch(range.at(0)) {
            case 'S':   // S<frame>: assume new scene at frame <frame>
                for (int i=startFrm; i<=endFrm; i++) {
                    SceneChgElement el = SceneChgElement(i, true);
                    sceneChgOverrides.insert(el);
                }
                break;
            case 'N':   // N<frame>: no new scenes in at <frame> or in range <startFrm> to <endFrm>
                for (int i=startFrm; i<=endFrm; i++) {
                    SceneChgElement el = SceneChgElement(i, false);
                    sceneChgOverrides.insert(el);
                }
                break;
            case 'E':   // E<frame>: leave frame <frame> or range <startFrm> until <endFrm> unchanged
                excludeFrames.push_back(IntPair(startFrm, endFrm));
                break;
            default:
                env->ThrowError("autolevels: bogus excludeframes specifier");
                break;
        }
    }

    greyscale_mode = luma_Rec601;
    if (strlen(matrix) > 0) {
        if (!vi.IsRGB())
            env->ThrowError("autolevels: invalid \"matrix\" parameter (RGB data only)");
        if (!lstrcmpi(matrix, "rec709"))
            greyscale_mode = luma_Rec709;
        else if (!lstrcmpi(matrix, "Average"))
            greyscale_mode = luma_avg;
        else if (!lstrcmpi(matrix, "rec601"))
            greyscale_mode = luma_Rec601;
        else
            env->ThrowError("GreyScale: invalid \"matrix\" parameter (must be matrix=\"Rec601\", \"Rec709\" or \"Average\")");
    }
}

PVideoFrame __stdcall Autolevels::GetFrame(int frameno, IScriptEnvironment* env)
{
    // check if the image is to be processed
    for (int i=0; i<(int)excludeFrames.size(); i++) {
        if (frameno >= excludeFrames[i].first && frameno <= excludeFrames[i].second)
            return child->GetFrame(frameno, env);   // return it unchanged
    }

    // range [avgFrom..avgTo] for determining the average of the ymin- and ymax-values
    int avgFrom = max(0, frameno-filterRadius);
    int avgTo = min(vi.num_frames-1, frameno+filterRadius);

    // restrict the range by scene changes found between avgFrom and avgTo
    boolean this_newscene = false;
    for (int i=avgFrom+1; i<=avgTo; i++) {
        boolean sceneChange = (i > 1) && isSceneStart(i, env);
        if (sceneChange) {
            if (i <= frameno) {
                avgFrom = i;
                this_newscene |= (i == frameno);
            } else {
                avgTo = i - 1;
                break;
            }
        }
    }

    // average of the minimum and maximum luma values
    IntPair yMinMaxAccum = IntPair(0, 0);
    int this_ymin = 0, this_ymax = 0;
    for (int i=avgFrom; i<=avgTo; i++) {
        IntPair yMinMaxFrame = getMinMax(i, env);
        yMinMaxAccum.first  += yMinMaxFrame.first;   // ymin
        yMinMaxAccum.second += yMinMaxFrame.second;  // ymax
        if (i == frameno) {
            this_ymin = yMinMaxFrame.first;
            this_ymax = yMinMaxFrame.second;
        }
    }
    double ymin = (double)yMinMaxAccum.first  / (avgTo-avgFrom+1);
    double ymax = (double)yMinMaxAccum.second / (avgTo-avgFrom+1);

    PVideoFrame frame = child->GetFrame(frameno, env);
    env->MakeWritable(&frame);

    if (ymax > ymin) {

        // map the luma values from the range of [ymin, ymax] to the range [16, 235]
        int yLUT[256];
        double yfac = 220.0 / (ymax-ymin);   // yfac and yoffs = luma scaling factor and offset
        double yoffs = -ymin * yfac + 16;
        for (int y=0; y<256; y++) {
            int yNew = (int)floor(y*yfac + yoffs + 0.5);
            yLUT[y] = min(max(yNew, 0), 255);   // clamp
        }

        // modify luma in place
        if (vi.IsYUY2()) {
            RemapYUY2(frame, yLUT);
        } else if (vi.IsPlanar() || vi.IsYUV()) {
            RemapPlanarY(frame, yLUT);
        } else if (!vi.IsPlanar() && vi.IsRGB()) {
            RemapInterleavedRGB(frame, yLUT);
        }

        if (debug) {
            sprintf(msgbuf, "ymin=%d, ymax=%d, ymin_avg=%5.1f, ymax_avg=%5.1f",
                     this_ymin, this_ymax, ymin, ymax);
            DrawString(vi, frame, 10, 20, msgbuf);
            if (this_newscene) {
                int color_orange = 10;
                DrawString(vi, frame, 10, 40, color_orange, "<<scene change>>");
            }
        }
    }

    return frame;
}

// This builds builds the limits of the histogram of the luma values for a
// given frame.  Instead of finding the true min/max luma, the darkest and
// brightest 1/256-th of all pixels are ignored, so that outliers don't
// overly influence the result.
IntPair Autolevels::getMinMax(int frameno, IScriptEnvironment* env)
{
    hash_map<int, IntPair>::const_iterator mmIter = MinMaxCache.find(frameno);
    int ymininner;
    int ymaxinner;

    if (mmIter == MinMaxCache.end()) {
        // compute the ymin/ymax values if not yet known

        int yhisto[256];
        for (int i=0; i<256; i++) {
            yhisto[i] = 0;
        }

        PVideoFrame frame = child->GetFrame(frameno, env);

        if (vi.IsYUY2()) {
            GetAvgLumaYUY2(frame, yhisto);
        } else if (vi.IsPlanar() || vi.IsYUV()) {
            GetAvgLumaPlanarY(frame, yhisto);
        } else if (!vi.IsPlanar() && vi.IsRGB()) {
            GetAvgLumaInterleavedRGB(frame, yhisto);
        }

        // find the limits of the distribution
        // JTB: I'm not sure why this is done, as the following code will skip
        //      over zero-valued buckets anyway, and this doesn't help speed.
        //      it might be useful in the future to know the absolute ymin/ymax.
        int ymin = 255;
        int ymax = 0;
        for (int y=0; y<256; y++) {
            if (yhisto[y] > 0) {
                ymin = (ymin == 255) ? y : ymin;
                ymax = y;
            }
        }

        // skip over the bottom and top 1/256-th quantile of the histogram
        // to find a more representative min/max value
        int numPixels256 = vi.width * vi.height / 256;
        ymininner = ymin;
        ymaxinner = ymax;

        int sumOutliers = yhisto[ymininner];
        while (sumOutliers < numPixels256) {
            ymininner++;
            sumOutliers += yhisto[ymininner];
        }

        sumOutliers = yhisto[ymaxinner];
        while (sumOutliers < numPixels256) {
            ymaxinner--;
            sumOutliers += yhisto[ymaxinner];
        }

        ymininner = max(ymininner, 16);   // set a lower bound for ymin and ...
        ymaxinner = min(ymaxinner, 235);  // ... upper bound for ymax

        IntPair ip = IntPair(ymininner, ymaxinner);
        MinMaxElement el = MinMaxElement(frameno, ip);
        MinMaxCache.insert(el);

    } else {
        // we've computed min/max before; just look it up
        IntPair ip = mmIter->second;
        ymininner = ip.first;
        ymaxinner = ip.second;
    }
    return IntPair(ymininner, ymaxinner);
}

// build the luma histogram for the frame
void Autolevels::GetAvgLumaPlanarY(PVideoFrame &frame, int *yhisto)
{
    const Pixel8 *srcp  = (Pixel8 *)frame->GetReadPtr(PLANAR_Y);
    const int src_pitch = frame->GetPitch(PLANAR_Y);

    for (int y=0; y<vi.height; y++) {
        for (int x=0; x<vi.width; x++) {
            yhisto[srcp[x]]++;
        }
        srcp += src_pitch;
    }
}

// build the luma histogram for the frame
void Autolevels::GetAvgLumaYUY2(PVideoFrame &frame, int *yhisto)
{
    const Pixel8 *srcp  = (Pixel8 *)frame->GetReadPtr();
    const int src_pitch = frame->GetPitch();

    for (int y=0; y<vi.height; y++) {
        const Pixel8 *pp = srcp;
        for (int x=0; x<vi.width; x++) {
            yhisto[*pp]++;
            pp += 2;
        }
        srcp += src_pitch;
    }
}

// build the luma histogram for the frame
//
// my understanding is that RGB packing has these byte offsets:
//    rgb24: blue = *(ptr+0), green = *(ptr+1), red = *(ptr+2)
//    rgb32: blue = *(ptr+0), green = *(ptr+1), red = *(ptr+2), alpha = *(ptr+3)
void Autolevels::GetAvgLumaInterleavedRGB(PVideoFrame &frame, int *yhisto)
{
    Pixel8   *srcp      = (Pixel8 *)frame->GetReadPtr();
    const int src_pitch = frame->GetPitch();
    const int rgb_inc   = (vi.IsRGB32()) ? 4 : 3;

    // modify in place
    for(int y=0; y<vi.height; y++) {
        Pixel8 *pp = srcp;
    // --- luma conversion stolen from avisynth/src/filters/greyscale.cpp ---
        switch (greyscale_mode) {

          case luma_Rec709: // This is the correct brigtness calculations (standardized in Rec. 709)
            for(int x=0; x<vi.width; x++) {
                int greyscale =((pp[0]*4732)+(pp[1]*46871)+(pp[2]*13933)+32768)>>16;
                yhisto[greyscale]++;
                pp += rgb_inc;
            }                                                                                   // at a time by temporarily treating the src and dst pointers as
            break;

          case luma_Rec601: // This produces similar results as YUY2 (luma calculation)
            for(int x=0; x<vi.width; x++) {
                int greyscale=((pp[0]*7471)+(pp[1]*38470)+(pp[2]*19595)+32768)>>16;
                yhisto[greyscale]++;
                pp += rgb_inc;
            }                                                                                   // at a time by temporarily treating the src and dst pointers as
            break;

          case luma_avg: // This is the unweighted average of R, G & B
            for(int x=0; x<vi.width; x++) {
                int greyscale=((pp[0]+pp[1]+pp[2])*21845+32768)>>16;
                yhisto[greyscale]++;
                pp += rgb_inc;
            }                                                                                   // at a time by temporarily treating the src and dst pointers as
            break;
        } // end switch
    // --- end ---

        srcp += src_pitch;
    }
}

// iterate over the image and remap the luminance channel
void Autolevels::RemapPlanarY(PVideoFrame &frame, int yLUT[])
{
    Pixel8   *dstp      = frame->GetWritePtr(PLANAR_Y);
    const int dst_pitch = frame->GetPitch(PLANAR_Y);

    for (int y=0; y<vi.height; y++) {
        Pixel8 *pp = dstp;
        for (int x=0; x<vi.width; x++) {
            *pp = yLUT[*pp];
            pp++;
        }
        dstp += dst_pitch;
    }
}

void Autolevels::RemapYUY2(PVideoFrame &frame, int yLUT[])
{
    Pixel8   *dstp      = frame->GetWritePtr();
    const int dst_pitch = frame->GetPitch();

    // modify in place
    for(int y=0; y<vi.height; y++) {
        Pixel8 *pp = dstp;
        for(int x=0; x<vi.width; x++) {
            *pp = yLUT[*pp];
            pp += 2;
        }                                                                                   // at a time by temporarily treating the src and dst pointers as
        dstp += dst_pitch;
    }
}

void Autolevels::RemapInterleavedRGB(PVideoFrame &frame, int yLUT[])
{
    Pixel8   *dstp      = frame->GetWritePtr();
    const int dst_pitch = frame->GetPitch();
    const int rgb_inc   = (vi.IsRGB32()) ? 4 : 3;

    // the proper thing to do would be to map each pixel from  rgb -> yuv,
    // then do y level remapping, then yuv -> rgb, along with some limit
    // checks and some risk of chroma shift if only one component gets clamped.
    //
    // instead we do something quick and dirty.  each r, g, b component will
    // run through the lut table as if each was a luma value.  blech.

    // modify in place
    for(int y=0; y<vi.height; y++) {
        Pixel8 *pp = dstp;
        for(int x=0; x<vi.width; x++) {
            pp[0] = yLUT[pp[0]];
            pp[1] = yLUT[pp[1]];
            pp[2] = yLUT[pp[2]];
            pp += rgb_inc;
        }
        dstp += dst_pitch;
    }
}

// This is not a scene change detection in the strict sense, but only checks
// whether the minimum or maximum luma changes greatly.  This simple approach
// is sufficient here, but perfect, for a scene change can be ignored without
// changing the minimum / maximum.
boolean Autolevels::isSceneStart(int frameno, IScriptEnvironment* env)
{
    hash_map<int, boolean>::const_iterator scgIter = sceneChgOverrides.find(frameno);
    if (scgIter != sceneChgOverrides.end()) {   // check if user overrides are in effect
        return scgIter->second;
    } else {
        if (frameno > 1) {
            IntPair minMax1 = getMinMax(frameno-1, env);
            IntPair minMax2 = getMinMax(frameno, env);
            int minDiff = abs(minMax2.first - minMax1.first);
            int maxDiff = abs(minMax2.second - minMax1.second);
            boolean sceneChange = (minDiff >= sceneChgThresh) || (maxDiff >= sceneChgThresh);
            return sceneChange;
        } else {
            return false;
        }
    }
}

AVSValue __cdecl Create_Autolevels(AVSValue args, void* user_data, IScriptEnvironment* env)
{
    return new Autolevels(args[0].AsClip(),
                          args[1].AsInt(filterRadiusDefault),
                          args[2].AsInt(sceneChgThreshDefault),
                          args[3].AsString(""),
                          args[4].AsString(""),
                          args[5].AsBool(debugDefault),
                          env
                         );
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("Autolevels",
                      // the argument string has been broken up,
                      // but note that it is actually one string!
                         "c"
                         "[filterRadius]i"
                         "[sceneChgThresh]i"
                         "[frameOverrides]s"
                         "[matrix]s"
                         "[debug]b",
                     Create_Autolevels, 0);
    return "`Autolevels' Histogram stretch plugin";
}

BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    return TRUE;
}

#ifdef _MANAGED
#pragma managed(pop)
#endif

